from typing import Annotated

from typer import Argument

from preling.app.app import app
from preling.db import get_path
from preling.utils.typer import typer_raise

__all__ = [
    'path',
]


@app.command()
def path(
        language: Annotated[
            str,
            Argument(help='Language code whose data file should be printed.'),
        ],
) -> None:
    """Print the absolute path to PreLing’s data file for `language`."""
    language_path = get_path(language)
    if language_path.exists():
        print(language_path.absolute())
    else:
        typer_raise(f'Language "{language}" is not initialized.')



from dataclasses import dataclass
from typing import Callable

from prompt_toolkit import PromptSession
from rich.console import Console
from rich.text import Text

from preling.utils.console import clear_current_line, clear_previous

__all__ = [
    'ask',
    'ExtraOption',
]

OPTION_DELIMITER = ' | '


@dataclass
class ExtraOption:
    title: str
    action: Callable[[], None]


def normalize_key(text: str) -> str:
    """Normalize the key for case-insensitive comparison."""
    return text.lower()


def build_options(extra_options: list[ExtraOption]) -> tuple[list[list[Text]], dict[str, ExtraOption]]:
    """Build the list of option prompts and a mapping from option keys to their actions."""
    prompt = []
    mapping: dict[str, ExtraOption] = {}
    for option in extra_options:
        if not option.title:
            raise ValueError('Option title cannot be empty.')
        prefix, suffix = option.title[0], option.title[1:]
        key = normalize_key(prefix)
        if key in mapping:
            raise ValueError(f'Duplicate key for options: {key}')
        mapping[key] = option
        prompt.append([Text(prefix, style='bold'), Text(suffix)])
    return prompt, mapping


def ask(
        console: Console,
        prompt: Text,
        extra_options: list[ExtraOption],
        *,
        on_before_input: Callable[[], None] | None = None,
) -> str:
    """Prompt the user for input or to choose one of the extra options."""
    options_prompt, options_mapping = build_options(extra_options)
    console.print(prompt, end='')
    for option_prompt in options_prompt:
        console.print(Text(OPTION_DELIMITER, style='dim'), end='')
        console.print(*option_prompt, sep='', end='')
    console.print()
    if on_before_input:
        on_before_input()
        clear_current_line()  # Otherwise the input made during the operation will get displayed twice
    while True:
        response = PromptSession().prompt()
        if option := options_mapping.get(normalize_key(response)):
            clear_previous()
            option.action()
        else:
            return response.strip()



from .delete import delete
from .init import init
from .path import path
from .stats import stats
from .study import study

__all__ = [
    'delete',
    'init',
    'path',
    'stats',
    'study',
]



from typing import Annotated

from typer import Argument, Option

from preling.app.app import app
from preling.db import get_path
from preling.utils.typer import typer_raise

__all__ = [
    'delete',
]


@app.command()
def delete(
        language: Annotated[
            str,
            Argument(help='Language code whose data should be removed.'),
        ],
        force: Annotated[
            bool | None,
            Option(
                '--force',
                '-f',
                help='Skip the confirmation prompt and delete immediately.',
            ),
        ] = False,
) -> None:
    """Delete all stored data for `language`."""
    path = get_path(language)
    if not path.exists():
        typer_raise(f'Language "{language}" is not initialized.')

    if not force and not input(
            f'Are you sure you want to delete all data for "{language}"? (y/N): ',
    ).lower().startswith('y'):
        typer_raise('Operation canceled.')

    path.unlink(missing_ok=True)
    print(f'Deleted all data for "{language}".')



from dataclasses import dataclass
import unicodedata

from openai import OpenAI
from pydantic import BaseModel

from preling.db import get_session
from preling.db.models import EvaluationCache, Sentence, Word

__all__ = [
    'evaluate',
    'SentenceEvaluation',
    'WordEvaluation',
]

TEMPERATURE = 0.0


@dataclass
class WordEvaluation:
    word: str
    word_data: Word
    is_correct: bool | None
    llm_translation: str


@dataclass
class SentenceEvaluation:
    is_correct: bool
    llm_translation: str
    back_translation: str | None
    words: list[WordEvaluation]


class WordOutput(BaseModel):
    word_number: int
    lowercased_word: str
    word: str
    llm_translation: str
    correctly_translated: bool | None


class SentenceOutput(BaseModel):
    llm_translation: str
    correctly_translated: bool
    back_translation: str | None
    words: list[WordOutput]


def normalize_spelling(word: str) -> str:
    return unicodedata.normalize('NFC', word.strip())


def generate_prompt(language: str, sentence: Sentence, translation: str) -> str:
    """
    Generate a prompt for the LLM to evaluate an English translation of a sentence
    from the specified source language.
    """
    return '\n'.join([
        f"A learner of the language with code `{language}` has attempted to translate "
        f"the following sentence from this language into English:",
        f"`{sentence.sentence}`",
        "Their translation is:",
        f"`{translation}`",
        "Determine whether the learner's translation is correct. Also, provide your own "
        "translation that you consider perfect. It may match the learner's version, or differ, "
        "even if you deem the learner's attempt correct but see room for minor improvements. "
        "If the learner's translation is incorrect, provide a back-translation of it into "
        "the original language.",
        "Additionally, here is the list of words in the sentence, in order (lowercased):",
        *[
            f"{word_number}) {word.word.lower()}"
            for word_number, word in enumerate(sentence.words, start=1)
        ],
        "For each word, provide the following information:",
        "- Your own translation of the word into English within the context of the sentence.",
        "- Whether the learner's translation of the word is correct—that is, whether all "
        "information conveyed by the word (lexical and/or grammatical) is present in the learner's "
        "translation. You may return null if it's unclear from the learner's translation whether "
        "they understood the word correctly.",
    ])


def build_evaluation(sentence_analysis: SentenceOutput, sentence: Sentence) -> SentenceEvaluation:
    """Build a `SentenceEvaluation` from the LLM's analysis of the sentence and its words."""
    return SentenceEvaluation(
        is_correct=sentence_analysis.correctly_translated,
        llm_translation=sentence_analysis.llm_translation,
        back_translation=sentence_analysis.back_translation,
        words=[
            WordEvaluation(
                word=word.word,
                word_data=sentence.words[word.word_number - 1],
                is_correct=word.correctly_translated,
                llm_translation=word.llm_translation,
            )
            for word in sentence_analysis.words
            if 1 <= word.word_number <= len(sentence.words)
            and (normalize_spelling(sentence.words[word.word_number - 1].word.lower()) ==
                 normalize_spelling(word.lowercased_word))
        ],
    )


def request(sentence: Sentence, language: str, translation: str, model: str, api_key: str) -> SentenceOutput:
    """Request the LLM to evaluate the translation of a sentence."""
    return OpenAI(api_key=api_key).responses.parse(
        model=model,
        temperature=TEMPERATURE,
        text_format=SentenceOutput,
        input=generate_prompt(language, sentence, translation),
    ).output_parsed


def cached_request(sentence: Sentence, language: str, translation: str, model: str, api_key: str) -> SentenceOutput:
    """Request the LLM to evaluate the translation of a sentence, using cache if available."""
    with get_session(language) as session:
        if cached := session.query(EvaluationCache).filter(
            EvaluationCache.sentence_id == sentence.id,
            EvaluationCache.model == model,
            EvaluationCache.translation == translation,
        ).first():
            return SentenceOutput.model_validate_json(cached.evaluation)
        else:
            output = request(sentence, language, translation, model, api_key)
            session.add(EvaluationCache(
                sentence_id=sentence.id,
                model=model,
                translation=translation,
                evaluation=output.model_dump_json(),
            ))
            return output


def evaluate(sentence: Sentence, language: str, translation: str, model: str, api_key: str) -> SentenceEvaluation:
    """Evaluate the translation of a sentence from the specified language into English using an LLM."""
    return build_evaluation(cached_request(sentence, language, translation, model, api_key), sentence)



from __future__ import annotations
from typing import Annotated

from dataclasses import dataclass, field
from rich.console import Console
from rich.text import Text
from tqdm import tqdm
from typer import Argument

from preling.app.app import app
from preling.db import get_session, Session
from preling.db.models import Sentence, Word
from preling.utils.time import get_timestamp

__all__ = [
    'stats',
]


@dataclass
class WordStatsItem:
    count: int = 0
    occurrences: int = 0

    def add_word(self, word: Word) -> None:
        """Add a word to the stats item."""
        self.count += 1
        self.occurrences += word.occurrences

    def format(self, total: int) -> Text:
        """Format the word stats item for display."""
        return Text(f'{self.count} ') + Text(f'(coverage: {self.occurrences / max(total, 1):.1%})', style='dim')


@dataclass
class WordStats:
    active: WordStatsItem = field(default_factory=WordStatsItem)
    seen: WordStatsItem = field(default_factory=WordStatsItem)
    total: WordStatsItem = field(default_factory=WordStatsItem)


@dataclass
class SentenceStats:
    correct: int = 0
    seen: int = 0
    total_sentences: int = 0
    total_attempts: int = 0


def compute_word_stats(session: Session) -> WordStats:
    """Compute word statistics for the current language."""
    now = get_timestamp()
    word_stats = WordStats()
    word: Word
    for word in tqdm(session.query(Word).all(), desc='Computing word statistics', leave=False):
        word_stats.total.add_word(word)
        if word.due is not None:
            word_stats.seen.add_word(word)
            if word.due > now:
                word_stats.active.add_word(word)
    return word_stats


def compute_sentence_stats(session: Session) -> SentenceStats:
    """Compute sentence statistics for the current language."""
    sentence_stats = SentenceStats()
    for sentence in tqdm(session.query(Sentence).all(), desc='Computing sentence statistics', leave=False):
        if sentence.correct_attempts:
            sentence_stats.correct += 1
        if sentence.incorrect_attempts or sentence.correct_attempts:
            sentence_stats.seen += 1
        sentence_stats.total_sentences += 1
        sentence_stats.total_attempts += sentence.correct_attempts + sentence.incorrect_attempts
    return sentence_stats


def format_section_title(title: str) -> Text:
    """Format a section title for display."""
    return Text(title, style='bold underline')


def format_stats_label(title: str) -> Text:
    """Format a stats title for display."""
    return Text(f'{title}:', style='bold')


def print_word_stats(console: Console, word_stats: WordStats) -> None:
    """Print word statistics to the console."""
    total_occurrences = word_stats.total.occurrences
    console.print(format_section_title('Word Statistics'))
    console.print(format_stats_label('In retention'), word_stats.active.format(total_occurrences))
    console.print(format_stats_label('Seen words'), word_stats.seen.format(total_occurrences))
    console.print(format_stats_label('Total words'), word_stats.total.format(total_occurrences))


def print_sentence_stats(console: Console, sentence_stats: SentenceStats) -> None:
    """Print sentence statistics to the console."""
    console.print(format_section_title('Sentence Statistics'))
    console.print(format_stats_label('Correct at least once'), sentence_stats.correct)
    console.print(
        format_stats_label('Seen sentences'),
        sentence_stats.seen,
        Text(f'(out of {sentence_stats.total_sentences})', style='dim'),
    )
    console.print(format_stats_label('Total attempts'), sentence_stats.total_attempts)


@app.command()
def stats(
        language: Annotated[
            str,
            Argument(help='Language code to show study statistics for.'),
        ],
) -> None:
    """Display study statistics for the given language."""
    with get_session(language) as session:
        word_stats = compute_word_stats(session)
        sentence_stats = compute_sentence_stats(session)

    console = Console(highlight=False)
    print_word_stats(console, word_stats)
    console.print()
    print_sentence_stats(console, sentence_stats)



from __future__ import annotations

from sqlalchemy import ForeignKey, Index, text
from sqlalchemy.orm import Mapped, mapped_column, relationship
from .base import Base

__all__ = [
    'EvaluationCache',
    'Sentence',
    'SentenceWord',
    'Word',
]


class Sentence(Base):
    __tablename__ = 'sentences'

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    sentence: Mapped[str]
    correct_attempts: Mapped[int]
    incorrect_attempts: Mapped[int]

    words: Mapped[list[Word]] = relationship(
        'Word',
        secondary='sentence_word_index',
        order_by=lambda: SentenceWord.word_index,
        back_populates='sentences',
        passive_deletes=True,
    )


class Word(Base):
    __tablename__ = 'words'

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    word: Mapped[str]
    occurrences: Mapped[int]
    streak_start: Mapped[int | None]  # Unix timestamp
    due: Mapped[int | None]  # Unix timestamp

    sentences: Mapped[list[Sentence]] = relationship(
        'Sentence',
        secondary='sentence_word_index',
        order_by=lambda: SentenceWord.random_key,
        back_populates='words',
        passive_deletes=True,
    )


Index('ix_words_id_due_null', Word.id, sqlite_where=Word.due.is_(None))
Index('ix_words_due_due_not_null', Word.due, sqlite_where=Word.due.is_not(None))


class SentenceWord(Base):
    __tablename__ = 'sentence_word_index'

    sentence_id: Mapped[int] = mapped_column(
        ForeignKey(Sentence.id, ondelete='CASCADE', onupdate='CASCADE'),
        primary_key=True,
    )
    word_index: Mapped[int] = mapped_column(primary_key=True)
    word_id: Mapped[int] = mapped_column(
        ForeignKey(Word.id, ondelete='CASCADE', onupdate='CASCADE'),
    )
    random_key: Mapped[int] = mapped_column(server_default=text('(abs(random()) % 16384)'))

    __table_args__ = (
        Index('ix_sentence_word_index_word_id_random_key', 'word_id', 'random_key'),
        {'sqlite_with_rowid': False},
    )


class EvaluationCache(Base):
    __tablename__ = 'evaluation_cache'

    sentence_id: Mapped[int] = mapped_column(
        ForeignKey(Sentence.id, ondelete='CASCADE', onupdate='CASCADE'),
        primary_key=True,
    )
    model: Mapped[str] = mapped_column(primary_key=True)
    translation: Mapped[str] = mapped_column(primary_key=True)
    evaluation: Mapped[str]



from __future__ import annotations
from collections import Counter
from pathlib import Path
from typing import Annotated, Generator, TYPE_CHECKING

if TYPE_CHECKING:
    from spacy.language import Language
from tqdm import tqdm
from typer import Argument

from preling.app.app import app
from preling.db import get_session, Session
from preling.db.models import Sentence, SentenceWord, Word
from preling.utils.typer import typer_raise

__all__ = [
    'init',
]


def get_nlp(language: str) -> Language:
    """Get a spaCy language model for the specified language."""
    import spacy
    try:
        return spacy.blank(language)
    except ImportError:
        typer_raise(f'Language "{language}" is not supported by spaCy.')


def get_sentences(corpus: Path) -> Generator[str, None, None]:
    """Yield sentences from the corpus file."""
    with corpus.open('r', encoding='utf-8') as file:
        for line in file:
            if sentence := line.strip():
                yield sentence


def extract_words(nlp: Language, sentence: str) -> list[str]:
    """Extract words from a sentence using spaCy."""
    return [token.lower_ for token in nlp(sentence) if any(c.isalpha() for c in token.text)]


def process_corpus(language: str, corpus: Path) -> tuple[dict[str, list[str]], Counter[str]]:
    """Process the corpus and return words by sentence and word frequencies."""
    nlp = get_nlp(language)

    words_by_sentence: dict[str, list[str]] = {}
    word_frequencies: Counter[str] = Counter()

    for sentence in tqdm(get_sentences(corpus), desc=f'Processing sentences'):
        if sentence not in words_by_sentence and (words := extract_words(nlp, sentence)):
            words_by_sentence[sentence] = words
            word_frequencies.update(words)

    return words_by_sentence, word_frequencies


def add_words(session: Session, frequencies: Counter[str]) -> dict[str, int]:
    """Add words to the database and return a mapping of words to their IDs."""
    ids_by_word: dict[str, int] = {}
    for word, occurrences in tqdm(frequencies.most_common(), desc='Adding words'):
        word_obj = Word(
            word=word,
            occurrences=occurrences,
            streak_start=None,
            due=None,
        )
        session.add(word_obj)
        session.flush()
        ids_by_word[word] = word_obj.id
    return ids_by_word


def add_sentences(session: Session, words_by_sentence: dict[str, list[str]], ids_by_word: dict[str, int]) -> None:
    """Add sentences to the database."""
    for sentence_text, words in tqdm(words_by_sentence.items(), desc='Adding sentences'):
        sentence_obj = Sentence(
            sentence=sentence_text,
            correct_attempts=0,
            incorrect_attempts=0,
        )
        session.add(sentence_obj)
        session.flush()
        for word_index, word in enumerate(words):
            session.add(SentenceWord(
                sentence_id=sentence_obj.id,
                word_index=word_index,
                word_id=ids_by_word[word],
            ))
        session.flush()


@app.command()
def init(
        language: Annotated[
            str,
            Argument(help='Language code supported by spaCy (e.g., "en", "fr", "uk").'),
        ],
        corpus: Annotated[
            Path,
            Argument(
                dir_okay=False,
                exists=True,
                readable=True,
                resolve_path=True,
                help='Plain‑text file containing one sentence per line.',
            ),
        ],
) -> None:
    """Initialise PreLing for a new language."""
    with get_session(language) as session:
        if session.query(Sentence).first():
            typer_raise(f'PreLing is already initialized for language "{language}".')
        words_by_sentence, word_frequencies = process_corpus(language, corpus)
        if not word_frequencies:
            typer_raise(f'No valid sentences found in the corpus.')
        ids_by_word = add_words(session, word_frequencies)
        add_sentences(session, words_by_sentence, ids_by_word)
        print('Committing changes to the database...')
        session.commit()
    print(f'Initialized PreLing for language "{language}" '
          f'with {len(words_by_sentence)} unique sentences '
          f'and {len(word_frequencies)} unique words.')



from contextlib import contextmanager
from pathlib import Path
from sqlite3 import Connection
from typing import Any, Generator

from sqlalchemy import create_engine, event, Engine
from sqlalchemy.orm import Session, sessionmaker

from preling.utils.paths import get_app_data_dir
from . import models  # Register models with SQLAlchemy
from .base import Base

__all__ = [
    'get_path',
    'get_session',
    'Session',
]

DB_FILENAME = '{language}.db'
SESSION_MAKERS: dict[str, sessionmaker[Session]] = {}


@event.listens_for(Engine, 'connect')
def set_sqlite_pragma(dbapi_connection: Connection, _connection_record: Any) -> None:
    """
    Enable foreign key support for SQLite.
    See https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#foreign-key-support
    """
    cursor = dbapi_connection.cursor()
    cursor.execute('PRAGMA foreign_keys=ON')
    cursor.close()


def get_path(language: str) -> Path:
    """Get the path to the database file for the specified language."""
    return get_app_data_dir() / DB_FILENAME.format(language=language)


@contextmanager
def get_session(language: str) -> Generator[Session, None, None]:
    if language not in SESSION_MAKERS:
        engine = create_engine(f'sqlite:///{get_path(language)}')
        Base.metadata.create_all(engine)
        SESSION_MAKERS[language] = sessionmaker(bind=engine)

    session = SESSION_MAKERS[language]()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()



import asyncio
from random import choice

from openai import AsyncOpenAI
from openai.helpers import LocalAudioPlayer

__all__ = [
    'read',
]

VOICES = [
    'alloy',
    'ash',
    'ballad',
    'coral',
    'echo',
    'fable',
    'nova',
    'onyx',
    'sage',
    'shimmer',
]

RESPONSE_FORMAT = 'pcm'


def read(text: str, language: str, model: str, api_key: str) -> None:
    """Read the given text using OpenAI's TTS service."""
    async def do_read() -> None:
        async with AsyncOpenAI(api_key=api_key) as client:
            async with client.audio.speech.with_streaming_response.create(
                    model=model,
                    voice=choice(VOICES),
                    input=text,
                    instructions=f'Language: {language}',
                    response_format=RESPONSE_FORMAT,
            ) as response:
                await LocalAudioPlayer().play(response)

    asyncio.run(do_read())



from openai import OpenAI
from typing import Generator

from preling.db.models import Sentence

__all__ = [
    'explain',
]

TEMPERATURE = 0.0


def generate_prompt(language: str, sentence: Sentence) -> str:
    """Generate a prompt instructing the LLM to explain a sentence written in a given language."""
    return '\n'.join([
        f"Here is a sentence written in the language with code `{language}`:",
        f"",
        f"`{sentence.sentence}`",
        f"",
        f"Explain to the learner of the language what this sentence means and how it is constructed. "
        f"Do not mention the language code `{language}` in your explanation.",
    ])


def explain(sentence: Sentence, language: str, model: str, api_key: str) -> Generator[str, None, None]:
    """Stream an explanation for a sentence written in a specified language using the LLM."""
    for event in OpenAI(api_key=api_key).responses.create(
        model=model,
        temperature=TEMPERATURE,
        input=generate_prompt(language, sentence),
        stream=True,
    ):
        if event.type == 'response.output_text.delta':
            yield event.delta



from datetime import timedelta

from preling.db.models import Sentence, Word
from preling.utils.time import get_timestamp

__all__ = [
    'update_sentence',
    'update_word',
]

INITIAL_INTERVAL = timedelta(hours=1)


def update_sentence(sentence: Sentence, is_correct: bool) -> None:
    """Update the sentence statistics based on whether the answer was correct."""
    if is_correct:
        sentence.correct_attempts += 1
    else:
        sentence.incorrect_attempts += 1


def update_word(word: Word, is_correct: bool) -> None:
    """Update spaced-repetition fields for a word depending on answer correctness."""
    now = get_timestamp()
    if is_correct:
        word.streak_start = word.streak_start or now
        word.due = now + max(now - word.streak_start, int(INITIAL_INTERVAL.total_seconds()))
    else:
        word.streak_start = None
        word.due = now



from functools import partial
from typing import Annotated

from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.text import Text
from typer import Argument, Option

from preling.app.app import app
from preling.db import get_session, Session
from preling.db.models import Sentence, Word
from preling.utils.console import clear_previous
from .chooser import choose_next
from .evaluator import evaluate, SentenceEvaluation, WordEvaluation
from .explainer import explain
from .interaction import ask, ExtraOption
from .tts import read
from .updater import update_sentence, update_word

__all__ = [
    'study',
]

SECTION_DELIMITER = Text('-' * 3, style='dim')

NEW_WORDS_SINGULAR = 'New word:'
NEW_WORDS_PLURAL = 'New words:'

AUDIO_ONLY_PROMPT = Text('(...)')

REPEAT_TITLE = 'Repeat'
EXPLAIN_TITLE = 'Explain'
QUIT_TITLE = 'Quit'

EVALUATION_IN_PROGRESS = Text('Evaluating translation...', style='dim')
ENTER_TO_CONTINUE = Text('Press Enter to continue...', style='dim')

YOURS = 'Yours:'
LLMS = "LLM's:"
BACK_TRANSLATION_DELIMITER = ' ~ '
CORRECT_MARK = '✅'
INCORRECT_MARK = '❌'
UNCLEAR_MARK = '❓'
LLM_MARK = '🔹'

CONSOLE = Console(highlight=False)

QUIT_OPTION = [ExtraOption(QUIT_TITLE, lambda: exit(0))]


def find_new_words(sentence: Sentence) -> dict[int, Word]:
    """Find words in a sentence that are not yet learned, indexed by their IDs."""
    return {word.id: word for word in sentence.words if word.due is None}


def print_new_words(words: dict[int, Word], target_word_id: int) -> None:
    """Print the new words found in a sentence, highlighting the target word."""
    CONSOLE.print(Text(NEW_WORDS_SINGULAR if len(words) == 1 else NEW_WORDS_PLURAL, style='yellow'))
    for word_id, word in sorted(words.items()):
        CONSOLE.print(
            Text(f'#{word_id}: ', style='red' if word_id == target_word_id else '')
            + Text(word.word.upper(), style='bold'),
        )
    CONSOLE.print(SECTION_DELIMITER)


def print_evaluation(translation: str, evaluation: SentenceEvaluation) -> None:
    """Print the evaluation results of a sentence translation."""
    CONSOLE.print(SECTION_DELIMITER)
    CONSOLE.print(' '.join([YOURS, CORRECT_MARK if evaluation.is_correct else INCORRECT_MARK, translation]), end='')
    if not evaluation.is_correct and evaluation.back_translation:
        CONSOLE.print(
            Text(BACK_TRANSLATION_DELIMITER, style='dim')
            + Text(evaluation.back_translation, style='bold'),
            end='',
        )
    CONSOLE.print()
    CONSOLE.print(' '.join([LLMS, LLM_MARK, evaluation.llm_translation]))
    CONSOLE.print(SECTION_DELIMITER)
    for e in evaluation.words:
        CONSOLE.print(
            Text(f'{CORRECT_MARK if e.is_correct else UNCLEAR_MARK if e.is_correct is None else INCORRECT_MARK} ')
            + Text(e.word, style='bold')
            + Text(f': {e.llm_translation}'),
        )


def update(sentence: Sentence, evaluation: SentenceEvaluation, session: Session) -> None:
    """Update the database with the results of the evaluation."""
    update_sentence(sentence, evaluation.is_correct)
    for e in evaluation.words:
        if e.is_correct is not None:
            update_word(e.word_data, e.is_correct)
    session.commit()


def build_explanation_option(sentence: Sentence, language: str, model: str, api_key: str) -> list[ExtraOption]:
    """Build an option to explain the sentence using the LLM."""
    def do_explain():
        chunks: list[str] = []
        with CONSOLE.screen():
            CONSOLE.print(sentence.sentence, style='bold red')
            CONSOLE.print()
            with Live(console=CONSOLE) as live:
                for chunk in explain(sentence, language, model, api_key):
                    chunks.append(chunk)
                    live.update(Markdown(''.join(chunks)))
            CONSOLE.print()
            CONSOLE.print(ENTER_TO_CONTINUE)
            input()

    return [ExtraOption(EXPLAIN_TITLE, do_explain)]


@app.command()
def study(
        model: Annotated[
            str,
            Option(
                '--model',
                envvar='PRELING_MODEL',
                help='GPT model for grammar evaluation.',
            ),
        ],
        tts_model: Annotated[
            str,
            Option(
                '--tts-model',
                envvar='PRELING_TTS_MODEL',
                help='Text‑to‑speech model.',
            ),
        ],
        api_key: Annotated[
            str,
            Option(
                '--api-key',
                envvar='PRELING_API_KEY',
                help='OpenAI API key.',
            ),
        ],
        language: Annotated[
            str,
            Argument(
                help='Language code previously initialised with `preling init`.',
            ),
        ],
        audio: Annotated[
            bool,
            Option(
                '--audio',
                help='Play audio in addition to printing text.',
            ),
        ] = False,
        audio_only: Annotated[
            bool,
            Option(
                '--audio-only',
                help='Play audio without displaying the text.',
            ),
        ] = False,
) -> None:
    """Launch an interactive study session."""
    with get_session(language) as session:
        target_word: Word
        sentence: Sentence

        def update_next() -> None:
            nonlocal target_word, sentence
            target_word, sentence = choose_next(session)

        update_next()

        while True:
            if new_words := find_new_words(sentence):
                print_new_words(new_words, target_word.id)
            formatted_sentence = Text(sentence.sentence, style='bold')

            do_read = partial(read, sentence.sentence, language, tts_model, api_key) if audio or audio_only else None
            repeat_option = [ExtraOption(REPEAT_TITLE, do_read)] if do_read else []

            translation = ask(
                CONSOLE,
                formatted_sentence if not audio_only else AUDIO_ONLY_PROMPT,
                repeat_option + QUIT_OPTION,
                on_before_input=do_read,
            )
            clear_previous(2)
            CONSOLE.print(formatted_sentence)
            CONSOLE.print(translation)
            CONSOLE.print(EVALUATION_IN_PROGRESS)
            evaluation = evaluate(sentence, language, translation, model, api_key)
            clear_previous()
            print_evaluation(translation, evaluation)
            update(sentence, evaluation, session)
            ask(
                CONSOLE,
                ENTER_TO_CONTINUE,
                repeat_option + build_explanation_option(sentence, language, model, api_key) + QUIT_OPTION,
                on_before_input=update_next,
            )
            clear_previous(2)
            CONSOLE.print(SECTION_DELIMITER)



from collections.abc import Iterable
from functools import partial
from random import random
from typing import cast

from preling.db import Session
from preling.db.models import Sentence, SentenceWord, Word
from preling.utils.time import get_timestamp
from preling.utils.typer import typer_raise

__all__ = [
    'choose_next',
]

MAX_CANDIDATE_SENTENCES = 1000  # Maximum number of rows to scan when choosing a sentence


def choose_next_word(session: Session) -> Word:
    """
    Return the next word to learn or review, using the following priority:
    1. The word with the earliest `due` timestamp that is already due.
    2. If no such word exists, the most frequent unlearned word (i.e., the one with the smallest `id` and `due`==None).
    3. If no unlearned words remain, the word with the earliest future `due`.
    """
    if next_due_word := session.query(Word).filter(
        Word.due.is_not(None),
        Word.due <= get_timestamp(),
    ).order_by(Word.due).first():
        return cast(Word, next_due_word)
    if next_most_frequent_word := session.query(Word).filter(
        Word.due.is_(None),
    ).order_by(Word.id).first():
        return cast(Word, next_most_frequent_word)
    if next_due_word := session.query(Word).filter(
        Word.due.is_not(None),
    ).order_by(Word.due).first():
        return cast(Word, next_due_word)
    typer_raise('No words to learn or review found in the database.')


def make_sentence_key(now: int, sentence: Sentence) -> tuple[int, int, int, float]:
    """
    Compute a ranking key for `sentence`:
    - the number of unlearned words in the sentence;
    - the *negative* number of due words in the sentence (so that `min()` maximizes them);
    - the sum of ids of unlearned words in the sentence;
    - a random float for tie‑breaking.
    """
    unlearned_count = sum(1 for word in sentence.words if word.due is None)
    due = sum(1 for word in sentence.words if word.due is not None and word.due <= now)
    unlearned_ids_sum = sum(word.id for word in sentence.words if word.due is None)
    return unlearned_count, -due, unlearned_ids_sum, random()


def choose_next_sentence(session: Session, word: Word) -> Sentence:
    """
    Choose a sentence for `word`, limited to the first `MAX_CANDIDATE_SENTENCES` matches, optimizing for:
    1. Fewest unlearned words.
    2. Most words that are currently due.
    3. Randomness among ties.
    """
    return min(
        (cast(
            Iterable[Sentence],
            session.query(Sentence)
            .join(SentenceWord, Sentence.id == SentenceWord.sentence_id)
            .filter(SentenceWord.word_id == word.id)
            .order_by(SentenceWord.random_key)
            .limit(MAX_CANDIDATE_SENTENCES),
        )),
        key=partial(make_sentence_key, get_timestamp()),
    )


def choose_next(session: Session) -> tuple[Word, Sentence]:
    """Pick the next word and sentence to learn or review."""
    word = choose_next_word(session)
    sentence = choose_next_sentence(session, word)
    return word, sentence
